from typing import Optional

from pydantic import Field, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

DEFAULT_SCOPE = "https://analysis.windows.net/powerbi/api/.default"
POWERBI_ENV_PREFIX = "CASTOR_POWERBI_"


class PowerbiCredentials(BaseSettings):
    """Class to handle PowerBI rest API permissions"""

    model_config = SettingsConfigDict(
        env_prefix=POWERBI_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )

    client_id: str
    tenant_id: str
    secret: str = Field(repr=False)
    scopes: list[str] = [DEFAULT_SCOPE]

    @field_validator("scopes", mode="before")
    @classmethod
    def _check_scopes(cls, scopes: Optional[list[str]]) -> list[str]:
        return scopes if scopes is not None else [DEFAULT_SCOPE]



from typing import Union

from tableauserverclient import ServerResponseError  # type: ignore
from typing_extensions import Literal

from .errors import TableauErrorCode

PageReturn = Union[
    tuple[list[dict], Literal[None]],
    tuple[Literal[None], Union[TableauErrorCode, ServerResponseError]],
]



from typing import Iterable, Optional

from .enums import LineageEntity


class LineageProcessor:
    """
    helper class that handles lineage deduplication and filtering
    """

    def __init__(self, lineage_entity: LineageEntity):
        self.lineage_entity = lineage_entity

        self.lineage: dict[tuple[str, str], dict] = dict()

    def _parent_path(self, link) -> Optional[str]:
        if self.lineage_entity == LineageEntity.TABLE:
            return link["source_table_full_name"]

        source_table = link["source_table_full_name"]
        source_column = link["source_column_name"]
        if not (source_table and source_column):
            return None

        return f"{source_table}.{source_column}"

    def _child_path(self, link) -> Optional[str]:
        if self.lineage_entity == LineageEntity.TABLE:
            return link["target_table_full_name"]

        target_table = link["target_table_full_name"]
        target_column = link["target_column_name"]
        if not (target_table and target_column):
            return None

        return f"{target_table}.{target_column}"

    def add(self, link: dict) -> None:
        """
        If the parent and child paths are valid, keeps the most recent lineage
        link in the `self.lineage` map.
        """
        parent = self._parent_path(link)
        child = self._child_path(link)
        timestamp = link["event_time"]

        if not (parent and child and parent != child):
            return

        key = (parent, child)
        if key in self.lineage and self.lineage[key]["event_time"] > timestamp:
            return

        self.lineage[key] = link


def valid_lineage(
    lineage: Iterable[dict], lineage_entity: LineageEntity
) -> list[dict]:
    """
    Filters out self-lineage or lineage with a missing source or target path,
    then deduplicates by picking the link with the most recent event timestmap.
    """
    deduplicated_lineage = LineageProcessor(lineage_entity)

    for link in lineage:
        deduplicated_lineage.add(link)

    return list(deduplicated_lineage.lineage.values())



from .utils import (
    usage_liveboard_reader,
)

VALID_CSV_1 = '''"Data extract produced by Castor on 09/19/2024 06:54"
"Filters applied on data :"
"User Action IN [pinboard_embed_view,pinboard_tspublic_no_runtime_filter,pinboard_tspublic_runtime_filter,pinboard_view]"
"Pinboard NOT IN [mlm - availability pinboard,null]"
"Timestamp >= 20240820 00:00:00 < 20240919 00:00:00"
"Timestamp >= 20240919 00:00:00 < 20240920 00:00:00"
""
"Pinboard","Pinboard Views","Unique Number of User"
"Market Report","559","19"
"Retailer report","204","14"
"Second-hand market","72","6"
"September test","25","2"'''


VALID_CSV_2 = '''"Data extract produced by Castor on 01/07/2025 16:07"
"Filters applied on data :"
"Timestamp >= 20241208 00:00:00 < 20250107 00:00:00"
""
"Answer name","User name","Number of unique users","Count of object interactions"
"toto","tata","1","666"'''

# Invalid CSV input (missing data rows)
INVALID_CSV = '''"Data extract produced by Castor on 09/19/2024 06:54"
"Filters applied on data :"
"User Action IN [pinboard_embed_view,pinboard_tspublic_no_runtime_filter,pinboard_tspublic_runtime_filter,pinboard_view]"
"Pinboard NOT IN [mlm - availability pinboard,null]"
"Timestamp >= 20240820 00:00:00 < 20240919 00:00:00"
"Timestamp >= 20240919 00:00:00 < 20240920 00:00:00"
""'''


def test_usage_liveboard_reader():
    expected_output_1 = [
        {
            "Pinboard": "Market Report",
            "Pinboard Views": "559",
            "Unique Number of User": "19",
        },
        {
            "Pinboard": "Retailer report",
            "Pinboard Views": "204",
            "Unique Number of User": "14",
        },
        {
            "Pinboard": "Second-hand market",
            "Pinboard Views": "72",
            "Unique Number of User": "6",
        },
        {
            "Pinboard": "September test",
            "Pinboard Views": "25",
            "Unique Number of User": "2",
        },
    ]
    expected_output_2 = [
        {
            "Answer name": "toto",
            "User name": "tata",
            "Number of unique users": "1",
            "Count of object interactions": "666",
        }
    ]

    result = list(usage_liveboard_reader(VALID_CSV_1))
    assert result == expected_output_1

    result = list(usage_liveboard_reader(VALID_CSV_2))
    assert result == expected_output_2

    result = list(usage_liveboard_reader(INVALID_CSV))
    assert result == []  # Expect an empty result since there is no data



import logging
from collections.abc import Iterable

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    write_json,
    write_summary,
)
from .assets import TableauRevampAsset
from .client import TableauRevampClient, TableauRevampCredentials

logger = logging.getLogger(__name__)


def iterate_all_data(
    client: TableauRevampClient,
) -> Iterable[tuple[TableauRevampAsset, list]]:
    """Iterate over the extracted Data from Tableau"""

    for asset in TableauRevampAsset:
        data = client.fetch(asset)
        yield asset, deep_serialize(data)


def extract_all(**kwargs) -> None:
    """
    Extract Data From tableau and store it locally in files under the
    output_directory
    """
    output_directory = kwargs.get("output") or from_env(OUTPUT_DIR)
    with_pulse = kwargs.get("with_pulse") or False
    timestamp = current_timestamp()

    credentials = TableauRevampCredentials(**kwargs)
    client = TableauRevampClient(credentials, with_pulse=with_pulse)
    client.login()

    for key, data in iterate_all_data(client):
        filename = get_output_filename(key.value, output_directory, timestamp)
        write_json(filename, data)

    write_summary(
        output_directory,
        timestamp,
        base_url=client.base_url,
        client_name=client.name,
    )



from collections.abc import Iterator
from functools import partial
from http import HTTPStatus
from typing import Optional

from ....utils import (
    APIClient,
    BasicAuth,
    RequestSafeMode,
    fetch_all_pages,
)
from ..assets import (
    ConfluenceAsset,
)
from .credentials import ConfluenceCredentials
from .endpoints import ConfluenceEndpointFactory
from .pagination import ConfluencePagination

_HEADERS = {
    "Accept": "application/json",
    "Content-Type": "application/json",
}

_MAX_ERROR_IGNORED_COUNT = 10
_IGNORED_ERROR_CODES = (HTTPStatus.BAD_GATEWAY,)
_SAFE_MODE = RequestSafeMode(
    max_errors=_MAX_ERROR_IGNORED_COUNT,
    status_codes=_IGNORED_ERROR_CODES,
)


class ConfluenceClient(APIClient):
    def __init__(
        self,
        credentials: ConfluenceCredentials,
        safe_mode: Optional[RequestSafeMode] = None,
    ):
        self.account_id = credentials.account_id
        auth = BasicAuth(
            username=credentials.username, password=credentials.password
        )
        super().__init__(
            auth=auth,
            host=credentials.base_url,
            headers=_HEADERS,
            safe_mode=safe_mode or _SAFE_MODE,
        )

    def pages(self):
        request = partial(
            self._get,
            endpoint=ConfluenceEndpointFactory.pages(),
            params={"body-format": "atlas_doc_format"},
        )
        yield from fetch_all_pages(request, ConfluencePagination)

    def users(self):
        request_body = {"accountIds": [self.account_id]}
        request = partial(
            self._post,
            endpoint=ConfluenceEndpointFactory.users(),
            data=request_body,
        )
        yield from fetch_all_pages(request, ConfluencePagination)

    def fetch(self, asset: ConfluenceAsset) -> Iterator[dict]:
        """Returns the needed metadata for the queried asset"""
        if asset == ConfluenceAsset.PAGES:
            yield from self.pages()

        elif asset == ConfluenceAsset.USERS:
            yield from self.users()

        else:
            raise ValueError(f"This asset {asset} is unknown")



from datetime import date, datetime
from typing import Optional

from ....utils import at_midnight, format_date, yesterday

_CLIENT_APP_BASE = "https://login.microsoftonline.com"
_REST_API_BASE_PATH = "https://api.powerbi.com/v1.0/myorg"


def _time_filter(day: Optional[date]) -> tuple[datetime, datetime]:
    target_day = day or yesterday()
    start = at_midnight(target_day)
    end = datetime.combine(target_day, datetime.max.time())
    return start, end


class PowerBiEndpointFactory:
    @classmethod
    def activity_events(cls, day: Optional[date]) -> str:
        start, end = _time_filter(day)
        url = f"{_REST_API_BASE_PATH}/admin/activityevents"
        url += "?$filter=Activity eq 'viewreport'"
        url += f"&startDateTime='{format_date(start)}'"
        url += f"&endDateTime='{format_date(end)}'"
        return url

    @classmethod
    def authority(cls, tenant_id: str) -> str:
        return f"{_CLIENT_APP_BASE}/{tenant_id}"

    @classmethod
    def dashboards(cls) -> str:
        return f"{_REST_API_BASE_PATH}/admin/dashboards"

    @classmethod
    def datasets(cls) -> str:
        return f"{_REST_API_BASE_PATH}/admin/datasets"

    @classmethod
    def groups(cls) -> str:
        return f"{_REST_API_BASE_PATH}/admin/groups"

    @classmethod
    def metadata_create_scan(cls) -> str:
        return f"{_REST_API_BASE_PATH}/admin/workspaces/getInfo"

    @classmethod
    def metadata_scan_result(cls, scan_id: int) -> str:
        return f"{_REST_API_BASE_PATH}/admin/workspaces/scanResult/{scan_id}"

    @classmethod
    def metadata_scan_status(cls, scan_id: int) -> str:
        return f"{_REST_API_BASE_PATH}/admin/workspaces/scanStatus/{scan_id}"

    @classmethod
    def pages(cls, report_id: str) -> str:
        return f"{_REST_API_BASE_PATH}/admin/reports/{report_id}/pages"

    @classmethod
    def reports(cls) -> str:
        return f"{_REST_API_BASE_PATH}/admin/reports"

    @classmethod
    def workspace_ids(cls) -> str:
        return f"{_REST_API_BASE_PATH}/admin/workspaces/modified"



from collections.abc import Iterator
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from http import HTTPStatus
from typing import Callable, Optional

import requests
from pydantic import BaseModel

from ....utils import (
    APIClient,
    BearerAuth,
    RequestSafeMode,
    build_url,
    fetch_all_pages,
    handle_response,
    retry,
)
from ..assets import SigmaAsset
from .credentials import SigmaCredentials
from .endpoints import SigmaEndpointFactory
from .pagination import SIGMA_API_LIMIT, SigmaPagination

_CONTENT_TYPE = "application/x-www-form-urlencoded"

_DATA_ELEMENTS: tuple[str, ...] = (
    "input-table",
    "pivot-table",
    "table",
    "visualization",
    "viz",
)

_AUTH_TIMEOUT_S = 60
_SIGMA_TIMEOUT_S = 300

_SIGMA_HEADERS = {
    "Content-Type": _CONTENT_TYPE,
}

_VOLUME_IGNORED = 10_000
_IGNORED_ERROR_CODES = (
    HTTPStatus.BAD_REQUEST,
    HTTPStatus.BAD_GATEWAY,
    HTTPStatus.INTERNAL_SERVER_ERROR,
    HTTPStatus.CONFLICT,
    HTTPStatus.NOT_FOUND,
)
SIGMA_SAFE_MODE = RequestSafeMode(
    max_errors=_VOLUME_IGNORED,
    status_codes=_IGNORED_ERROR_CODES,
)
_THREADS_LINEAGE = 10  # empirically found; hit the rate limit with 20 workers
_RETRY_NUMBER = 1
_RETRY_BASE_MS = 60_000


class LineageContext(BaseModel):
    """all info needed to build the endpoint for lineage retrieval"""

    workbook_id: str
    element_id: str


class Lineage(BaseModel):
    """holds response from lineage API and context used to retrieve it"""

    lineage: dict
    context: LineageContext


class SigmaBearerAuth(BearerAuth):
    def __init__(self, host: str, token_payload: dict[str, str]):
        auth_endpoint = SigmaEndpointFactory.authentication()
        self.authentication_url = build_url(host, auth_endpoint)
        self.token_payload = token_payload

    def fetch_token(self):
        token_api_path = self.authentication_url
        token_response = requests.post(
            token_api_path, data=self.token_payload, timeout=_AUTH_TIMEOUT_S
        )
        return handle_response(token_response)["access_token"]


class SigmaClient(APIClient):
    def __init__(
        self,
        credentials: SigmaCredentials,
        safe_mode: Optional[RequestSafeMode] = None,
    ):
        auth = SigmaBearerAuth(
            host=credentials.host,
            token_payload=credentials.token_payload,
        )
        super().__init__(
            host=credentials.host,
            auth=auth,
            headers=_SIGMA_HEADERS,
            timeout=_SIGMA_TIMEOUT_S,
            safe_mode=safe_mode or SIGMA_SAFE_MODE,
        )

    def _get_paginated(self, endpoint: str) -> Callable:
        return partial(
            self._get, endpoint=endpoint, params={"limit": SIGMA_API_LIMIT}
        )

    def _get_all_datasets(self) -> Iterator[dict]:
        request = self._get_paginated(endpoint=SigmaEndpointFactory.datasets())
        yield from fetch_all_pages(request, SigmaPagination)

    def _get_all_files(self) -> Iterator[dict]:
        request = self._get_paginated(endpoint=SigmaEndpointFactory.files())
        yield from fetch_all_pages(request, SigmaPagination)

    def _get_all_members(self) -> Iterator[dict]:
        request = self._get_paginated(endpoint=SigmaEndpointFactory.members())
        yield from fetch_all_pages(request, SigmaPagination)

    def _get_all_workbooks(self) -> Iterator[dict]:
        request = self._get_paginated(endpoint=SigmaEndpointFactory.workbooks())
        yield from fetch_all_pages(request, SigmaPagination)

    def _get_elements_per_page(
        self, page: dict, workbook_id: str
    ) -> Iterator[dict]:
        page_id = page["pageId"]
        request = self._get_paginated(
            SigmaEndpointFactory.elements(workbook_id, page_id)
        )
        elements = fetch_all_pages(request, SigmaPagination)
        for element in elements:
            if element.get("type") not in _DATA_ELEMENTS:
                continue
            yield {
                **element,
                "workbook_id": workbook_id,
                "page_id": page_id,
            }

    def _get_all_elements(self, workbooks: list[dict]) -> Iterator[dict]:
        for workbook in workbooks:
            workbook_id = workbook["workbookId"]

            request = self._get_paginated(
                SigmaEndpointFactory.pages(workbook_id)
            )
            pages = fetch_all_pages(request, SigmaPagination)

            for page in pages:
                yield from self._get_elements_per_page(
                    page=page, workbook_id=workbook_id
                )

    @retry(
        (ConnectionError,),
        max_retries=_RETRY_NUMBER,
        base_ms=_RETRY_BASE_MS,
        log_exc_info=True,
    )
    def _get_lineage(self, lineage_context: LineageContext) -> Lineage:
        """
        return the lineage from API and other ids needed to characterize
        lineage in castor
        """
        workbook_id = lineage_context.workbook_id
        element_id = lineage_context.element_id
        endpoint = SigmaEndpointFactory.lineage(workbook_id, element_id)
        return Lineage(lineage=self._get(endpoint), context=lineage_context)

    @staticmethod
    def _lineage_context(elements: list[dict]) -> list[LineageContext]:
        """
        Helper function to prepare context for lineage retrieval.
        Elements without associated columns are skipped.
        """
        contexts: list[LineageContext] = []
        for element in elements:
            if element.get("columns") is None:
                continue

            context = LineageContext(
                workbook_id=element["workbook_id"],
                element_id=element["elementId"],
            )
            contexts.append(context)
        return contexts

    def _get_all_lineages(self, elements: list[dict]) -> Iterator[dict]:
        lineage_context = self._lineage_context(elements)

        with ThreadPoolExecutor(max_workers=_THREADS_LINEAGE) as executor:
            results = executor.map(self._get_lineage, lineage_context)

        for lineage in results:
            yield {
                **lineage.lineage,
                "workbook_id": lineage.context.workbook_id,
                "element_id": lineage.context.element_id,
            }

    def _get_all_queries(self, workbooks: list[dict]) -> Iterator[dict]:
        for workbook in workbooks:
            workbook_id = workbook["workbookId"]
            request = self._get_paginated(
                SigmaEndpointFactory.queries(workbook_id)
            )
            queries = fetch_all_pages(request, SigmaPagination)

            for query in queries:
                yield {**query, "workbook_id": workbook_id}

    def fetch(
        self,
        asset: SigmaAsset,
        workbooks: Optional[list[dict]] = None,
        elements: Optional[list[dict]] = None,
    ) -> Iterator[dict]:
        """Returns the needed metadata for the queried asset"""
        if asset == SigmaAsset.DATASETS:
            yield from self._get_all_datasets()

        elif asset == SigmaAsset.ELEMENTS:
            if not workbooks:
                raise ValueError("Missing workbooks to extract elements")

            yield from self._get_all_elements(workbooks)

        elif asset == SigmaAsset.FILES:
            yield from self._get_all_files()

        elif asset == SigmaAsset.LINEAGES:
            if not elements:
                raise ValueError("Missing elements to extract lineage")
            yield from self._get_all_lineages(elements)

        elif asset == SigmaAsset.MEMBERS:
            yield from self._get_all_members()

        elif asset == SigmaAsset.QUERIES:
            if not workbooks:
                raise ValueError("Missing workbooks to extract queries")

            yield from self._get_all_queries(workbooks)

        elif asset == SigmaAsset.WORKBOOKS:
            yield from self._get_all_workbooks()

        else:
            raise ValueError(f"This asset {asset} is unknown")



SCOPES = (
    "https://www.googleapis.com/auth/datastudio",
    "https://www.googleapis.com/auth/userinfo.profile",
    "https://www.googleapis.com/auth/admin.reports.audit.readonly",
    "https://www.googleapis.com/auth/admin.directory.user.readonly",
)



from datetime import date
from unittest.mock import Mock, call, patch

import pytest

from .authentication import msal
from .client import PowerbiClient
from .constants import Keys
from .credentials import PowerbiCredentials
from .endpoints import PowerBiEndpointFactory

FAKE_TENANT_ID = "IamFake"
FAKE_CLIENT_ID = "MeTwo"
FAKE_SECRET = "MeThree"


@pytest.fixture
def mock_msal():
    with patch.object(msal, "ConfidentialClientApplication") as mock_app:
        mock_app.return_value.acquire_token_for_client.return_value = {
            "access_token": "fake_token"
        }
        yield mock_app


@pytest.fixture
def power_bi_client(mock_msal):
    creds = PowerbiCredentials(
        tenant_id=FAKE_TENANT_ID,
        client_id=FAKE_CLIENT_ID,
        secret=FAKE_SECRET,
    )
    return PowerbiClient(creds)


def test__access_token(power_bi_client, mock_msal):
    # Valid token scenario
    valid_token = "mock_token"
    mock_response = {"access_token": valid_token}
    returning_valid_token = Mock(return_value=mock_response)
    mock_msal.return_value.acquire_token_for_client = returning_valid_token

    assert power_bi_client._auth.fetch_token() == valid_token

    # Invalid token scenario
    invalid_response = {"not_access_token": "666"}
    returning_invalid_token = Mock(return_value=invalid_response)
    mock_msal.return_value.acquire_token_for_client = returning_invalid_token

    with pytest.raises(ValueError):
        power_bi_client._auth.fetch_token()


def test__datasets(power_bi_client):
    with patch.object(power_bi_client, "_get") as mocked_get:
        mocked_get.return_value = {"value": [{"id": 1, "type": "dataset"}]}
        datasets = list(power_bi_client._datasets())
        mocked_get.assert_called_with(PowerBiEndpointFactory.datasets())
        assert datasets == [{"id": 1, "type": "dataset"}]


def test__dashboards(power_bi_client):
    with patch.object(power_bi_client, "_get") as mocked_get:
        mocked_get.return_value = {"value": [{"id": 1, "type": "dashboard"}]}
        dashboards = list(power_bi_client._dashboards())
        mocked_get.assert_called_with(PowerBiEndpointFactory.dashboards())
        assert dashboards == [{"id": 1, "type": "dashboard"}]


def test__reports(power_bi_client):
    with patch.object(power_bi_client, "_get") as mocked_get:
        mocked_get.side_effect = [
            {"value": [{"id": 1, "type": "report"}]},
            {
                "value": [
                    {"name": "page_name", "displayName": "page", "order": 0}
                ]
            },
        ]
        reports = list(power_bi_client._reports())
        calls = [
            call(PowerBiEndpointFactory.reports()),
            call(PowerBiEndpointFactory.pages("1")),
        ]
        mocked_get.assert_has_calls(calls)
        assert reports == [
            {
                "id": 1,
                "type": "report",
                "pages": [
                    {"name": "page_name", "displayName": "page", "order": 0}
                ],
            }
        ]


def test__workspace_ids(power_bi_client):
    with patch.object(power_bi_client, "_get") as mocked_get:
        mocked_get.return_value = [{"id": 1000}, {"id": 1001}, {"id": 1003}]

        ids = power_bi_client._workspace_ids()
        assert ids == [1000, 1001, 1003]

        params = {
            Keys.INACTIVE_WORKSPACES: True,
            Keys.PERSONAL_WORKSPACES: True,
        }

        mocked_get.assert_called_with(
            PowerBiEndpointFactory.workspace_ids(),
            params=params,
        )


@patch.object(PowerbiClient, "_get_scan_result")
@patch.object(PowerbiClient, "_wait_for_scan_result")
@patch.object(PowerbiClient, "_create_scan")
@patch.object(PowerbiClient, "_workspace_ids")
def test__metadata(
    mock_workspace_ids,
    mock_create_scan,
    mock_wait_for_scan,
    mock_get_scan_result,
    power_bi_client,
):
    mock_workspace_ids.return_value = list(range(200))
    mock_create_scan.return_value = 314
    mock_wait_for_scan.return_value = True
    mock_get_scan_result.return_value = [{"workspace_id": 1871}]

    result = list(power_bi_client._metadata())

    assert result == [{"workspace_id": 1871}, {"workspace_id": 1871}]


def test__activity_events(power_bi_client):
    day = date.today()
    mocked_get_results = [
        {
            Keys.ACTIVITY_EVENT_ENTITIES: ["foo", "bar"],
            Keys.LAST_RESULT_SET: False,
            Keys.CONTINUATION_URI: "https://next-call-1",
        },
        {
            Keys.ACTIVITY_EVENT_ENTITIES: ["baz"],
            Keys.LAST_RESULT_SET: False,
            Keys.CONTINUATION_URI: "https://next-call-2",
        },
        {
            Keys.ACTIVITY_EVENT_ENTITIES: ["biz"],
            Keys.LAST_RESULT_SET: True,
            Keys.CONTINUATION_URI: None,
        },
    ]

    with patch.object(power_bi_client, "_get") as mocked_get:
        mocked_get.side_effect = mocked_get_results

        result = list(power_bi_client._activity_events(day=day))
        assert result == ["foo", "bar", "baz", "biz"]

        expected_calls = [
            call(endpoint=PowerBiEndpointFactory.activity_events(day=day)),
            call(endpoint="https://next-call-1"),
            call(endpoint="https://next-call-2"),
        ]
        mocked_get.assert_has_calls(expected_calls)


def test_test_connection(power_bi_client):
    with patch.object(power_bi_client._auth, "refresh_token") as mock_refresh:
        power_bi_client.test_connection()
        mock_refresh.assert_called_once()



class ConfluenceEndpointFactory:
    """
    Confluence rest api v2 endpoint factory.
    https://developer.atlassian.com/cloud/confluence/rest/v2/intro/#about
    """

    API = "wiki/api/v2/"
    PAGES = "pages"
    USERS = "users-bulk"

    @classmethod
    def pages(cls) -> str:
        """
        Endpoint to fetch all pages.
        More: https://developer.atlassian.com/cloud/confluence/rest/v2/api-group-page/#api-pages-get
        """
        return f"{cls.API}{cls.PAGES}"

    @classmethod
    def users(cls) -> str:
        """
        Endpoint to fetch all user.
        More: https://developer.atlassian.com/cloud/confluence/rest/v2/api-group-user/#api-users-bulk-post
        """
        return f"{cls.API}{cls.USERS}"



from typing import Iterator

from .. import LookerStudioAsset
from .admin_sdk_client import USER_EMAIL_FIELD, AdminSDKClient
from .credentials import LookerStudioCredentials
from .looker_studio_api_client import LookerStudioAPIClient


class LookerStudioClient:
    """
    Acts as a wrapper class to fetch Looker Studio assets, which requires
    coordinating calls between the Admin SDK API and the Looker Studio API.
    """

    def __init__(self, credentials: LookerStudioCredentials):
        self.admin_sdk_client = AdminSDKClient(credentials)
        self.looker_studio_client = LookerStudioAPIClient(credentials)

    def _get_assets(self) -> Iterator[dict]:
        """
        Extracts reports and data sources user by user.
        """
        users = self.admin_sdk_client.list_users()

        for user in users:
            email = user[USER_EMAIL_FIELD]
            yield from self.looker_studio_client.fetch_user_assets(email)

    def fetch(self, asset: LookerStudioAsset) -> Iterator[dict]:
        if asset == LookerStudioAsset.VIEW_ACTIVITY:
            yield from self.admin_sdk_client.list_view_events()

        elif asset == LookerStudioAsset.ASSETS:
            yield from self._get_assets()

        else:
            raise ValueError(f"The asset {asset}, is not supported")



from collections.abc import Iterator
from typing import Optional

import tableauserverclient as TSC  # type: ignore

from ....utils import SerializedAsset, retry
from ..assets import TableauRevampAsset
from ..constants import DEFAULT_PAGE_SIZE
from .errors import TableauApiError, TableauApiTimeout
from .gql_queries import FIELDS_QUERIES, GQL_QUERIES, QUERY_TEMPLATE

# increase the value when extraction is too slow
# decrease the value when timeouts arise
_CUSTOM_PAGE_SIZE: dict[TableauRevampAsset, int] = {
    # for some clients, extraction of columns tend to hit the node limit
    # https://community.tableau.com/s/question/0D54T00000YuK60SAF/metadata-query-nodelimitexceeded-error
    # the workaround is to reduce pagination
    TableauRevampAsset.COLUMN: 50,
    # fields are light but volumes are bigger
    TableauRevampAsset.FIELD: 1000,
    TableauRevampAsset.TABLE: 50,
}

_TIMEOUT_MESSAGE = (
    "Execution canceled because timeout of 30000 millis was reached"
)

_RETRY_BASE_MS = 10_000
_RETRY_COUNT = 4


def _check_errors(answer: dict) -> None:
    """
    handle errors in graphql response:
    - return None when there's no errors in the answer
    - TableauApiTimeout if any of the errors is a timeout
    - TableauApiError (generic) otherwise
    """
    if "errors" not in answer:
        return

    errors = answer["errors"]

    for error in errors:
        if error.get("message") == _TIMEOUT_MESSAGE:
            # we need specific handling for timeout issues (retry strategy)
            raise TableauApiTimeout(errors)

    raise TableauApiError(answer["errors"])


def gql_query_scroll(
    server,
    query: str,
    resource: str,
) -> Iterator[SerializedAsset]:
    """
    Iterate over GQL query results, handling pagination and cursor

    We have a retry strategy when timeout issues arise.
    It's a known issue on Tableau side, still waiting for their fix:
    https://issues.salesforce.com/issue/a028c00000zKahoAAC/undefined
    """

    @retry(
        exceptions=(TableauApiTimeout,),
        max_retries=_RETRY_COUNT,
        base_ms=_RETRY_BASE_MS,
    )
    def _call(cursor: Optional[str]) -> dict:
        # If cursor is defined it must be quoted else use null token
        token = "null" if cursor is None else f'"{cursor}"'
        query_ = query.replace("AFTER_TOKEN_SIGNAL", token)
        answer = server.metadata.query(query_)
        _check_errors(answer)
        return answer["data"][f"{resource}Connection"]

    cursor = None
    while True:
        payload = _call(cursor)
        yield payload["nodes"]

        page_info = payload["pageInfo"]
        if page_info["hasNextPage"]:
            cursor = page_info["endCursor"]
        else:
            break


class TableauClientMetadataApi:
    """
    Calls the MetadataAPI, using graphQL
    https://help.tableau.com/current/api/metadata_api/en-us/reference/index.html
    """

    def __init__(
        self,
        server: TSC.Server,
        override_page_size: Optional[int] = None,
    ):
        self._server = server
        self._forced_page_size = override_page_size

    def _call(
        self,
        resource: str,
        fields: str,
        page_size: int = DEFAULT_PAGE_SIZE,
    ) -> SerializedAsset:
        query = QUERY_TEMPLATE.format(
            resource=resource,
            fields=fields,
            page_size=page_size,
        )
        result_pages = gql_query_scroll(self._server, query, resource)
        return [asset for page in result_pages for asset in page]

    def _page_size(self, asset: TableauRevampAsset) -> int:
        return (
            self._forced_page_size
            or _CUSTOM_PAGE_SIZE.get(asset)
            or DEFAULT_PAGE_SIZE
        )

    def _fetch_fields(self) -> SerializedAsset:
        result: SerializedAsset = []
        page_size = self._page_size(TableauRevampAsset.FIELD)
        for resource, fields in FIELDS_QUERIES:
            current = self._call(resource, fields, page_size)
            result.extend(current)
        return result

    def fetch(
        self,
        asset: TableauRevampAsset,
    ) -> SerializedAsset:
        if asset == TableauRevampAsset.FIELD:
            return self._fetch_fields()

        page_size = self._page_size(asset)
        resource, fields = GQL_QUERIES[asset]
        return self._call(resource, fields, page_size)



import csv
import re
from collections.abc import Iterator
from io import StringIO

_END_OF_GENERATED_TEXT = r'^""$'


def usage_liveboard_reader(usage_liveboard_csv: str) -> Iterator[dict]:
    """
    Converts a CSV string into an iterator of dictionaries after
    ignoring the generated text that preceeds the actual CSV header row.
    The generated block ends with a row containing only two double quotes.
    Here is an example:

        "Data extract produced by Castor on 09/19/2024 06:54"
        "Filters applied on data :"
        "User Action IN [pinboard_embed_view,pinboard_tspublic_no_runtime_filter,pinboard_tspublic_runtime_filter,pinboard_view]"
        "Pinboard NOT IN [mlm - availability pinboard,null]"
        "Timestamp >= 20240820 00:00:00 < 20240919 00:00:00"
        "Timestamp >= 20240919 00:00:00 < 20240920 00:00:00"
        ""

    """
    csv_file = StringIO(usage_liveboard_csv)

    line = next(csv_file)
    while not re.match(_END_OF_GENERATED_TEXT, line.strip()):
        line = next(csv_file)

    yield from csv.DictReader(csv_file)



from pydantic import BaseModel, SecretStr, field_serializer


class LookerStudioCredentials(BaseModel):
    admin_email: str
    auth_provider_x509_cert_url: str
    auth_uri: str
    client_email: str
    client_id: str
    client_x509_cert_url: str
    private_key: SecretStr
    private_key_id: str
    project_id: str
    token_uri: str
    type: str

    @field_serializer("private_key")
    def dump_secret(self, pk):
        """When using model_dump, show private_key value"""
        return pk.get_secret_value()
